@@ -6,6 +6,8 @@ from scipy.stats import chi2, norm
 from causallearn.utils.KCI.KCI import KCI_CInd, KCI_UInd
 from causallearn.utils.PCUtils import Helper
 
+CONST_BINCOUNT_UNIQUE_THRESHOLD = 1e5
+
 
 def kci(data, X, Y, condition_set=None, kernelX='Gaussian', kernelY='Gaussian', kernelZ='Gaussian',
         est_width='empirical', polyd=2, kwidthx=None, kwidthy=None, kwidthz=None):
@@ -19,7 +21,7 @@ def kci(data, X, Y, condition_set=None, kernelX='Gaussian', kernelY='Gaussian',
 
 
 def kci_ui(X, Y, kernelX='Gaussian', kernelY='Gaussian', est_width='empirical', polyd=2, kwidthx=None, kwidthy=None):
-    '''
+    """
      To test if x and y are unconditionally independent
        Parameters
        ----------
@@ -34,7 +36,7 @@ def kci_ui(X, Y, kernelX='Gaussian', kernelY='Gaussian', est_width='empirical',
        polyd: polynomial kernel degrees (default=2)
        kwidthx: kernel width for data x (standard deviation sigma)
        kwidthy: kernel width for data y (standard deviation sigma)
-    '''
+    """
 
     kci_uind = KCI_UInd(kernelX, kernelY, est_width=est_width, polyd=polyd, kwidthx=kwidthx, kwidthy=kwidthy)
     pvalue, _ = kci_uind.compute_pvalue(X, Y)
@@ -43,7 +45,7 @@ def kci_ui(X, Y, kernelX='Gaussian', kernelY='Gaussian', est_width='empirical',
 
 def kci_ci(X, Y, Z, kernelX='Gaussian', kernelY='Gaussian', kernelZ='Gaussian', est_width='empirical', polyd=2,
            kwidthx=None, kwidthy=None, kwidthz=None):
-    '''
+    """
      To test if x and y are conditionally independent given z
        Parameters
        ----------
@@ -60,7 +62,7 @@ def kci_ci(X, Y, Z, kernelX='Gaussian', kernelY='Gaussian', kernelZ='Gaussian',
        kwidthx: kernel width for data x (standard deviation sigma, default None)
        kwidthy: kernel width for data y (standard deviation sigma)
        kwidthz: kernel width for data y (standard deviation sigma)
-    '''
+    """
 
     kci_cind = KCI_CInd(kernelX, kernelY, kernelZ, est_width=est_width, polyd=polyd, kwidthx=kwidthx, kwidthy=kwidthy,
                         kwidthz=kwidthz)
@@ -68,8 +70,8 @@ def kci_ci(X, Y, Z, kernelX='Gaussian', kernelY='Gaussian', kernelZ='Gaussian',
     return pvalue
 
 
-def mv_fisherz(mvdata, X, Y, condition_set, sample_size=None):
-    '''
+def mv_fisherz(mvdata, X, Y, condition_set):
+    """
     Perform an independence test using Fisher-Z's test for data with missing values
 
     Parameters
@@ -80,19 +82,15 @@ def mv_fisherz(mvdata, X, Y, condition_set, sample_size=None):
     Returns
     -------
     p : the p-value of the test
-    '''
+    """
     var = list((X, Y) + condition_set)
-    sub_corr_matrix, del_sample_size = get_sub_correlation_matrix(mvdata[:, var])  # the columns represent variables
-    sample_size = del_sample_size
-    inv = np.linalg.inv(sub_corr_matrix)
-    r = -inv[0, 1] / sqrt(inv[0, 0] * inv[1, 1])
-    Z = 0.5 * log((1 + r) / (1 - r))
-    X = sqrt(sample_size - len(condition_set) - 3) * abs(Z)
-    p = 1 - norm.cdf(abs(X))
-    return p
+    test_wise_deletion_XYcond_rows_index = get_index_no_mv_rows(mvdata[:, var])
+    assert len(test_wise_deletion_XYcond_rows_index) != 0, "A test-wise deletion fisher-z test appears no overlapping data of involved variables. Please check the input data."
+    test_wise_deleted_data = mvdata[test_wise_deletion_XYcond_rows_index,:]
+    return fisherz(test_wise_deleted_data, X, Y, condition_set)
 
 
-def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
+def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set):
     """Perform an independent test using Fisher-Z's test with test-wise deletion and missingness correction
     If it is not the case which requires a correction, then call function mvfisherZ(...)
     :param prt_m: dictionary, with elements:
@@ -102,7 +100,7 @@ def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
 
     ## Check whether whether there is at least one common child of X and Y
     if not Helper.cond_perm_c(X, Y, condition_set, prt_m, skel):
-        return mv_fisherz(mdata, X, Y, condition_set, sample_size)
+        return mv_fisherz(mdata, X, Y, condition_set)
 
     ## *********** Step 1 ***********
     # Learning generaive model for {X, Y, S} to impute X, Y, and S
@@ -112,7 +110,7 @@ def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
     W_indx_ = Helper.get_prt_mvars(var=list((X, Y) + condition_set), prt_m=prt_m)
 
     if len(W_indx_) == 0:  # When there is no variable can be used for correction
-        return mv_fisherz(mdata, X, Y, condition_set, sample_size)
+        return mv_fisherz(mdata, X, Y, condition_set)
 
     ## Get the parents of W missingness indicators
     W_indx = Helper.get_prt_mw(W_indx_, prt_m)
@@ -124,7 +122,7 @@ def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
     W_indx = list(set(W_indx) - set(var))
 
     if len(W_indx) == 0:  # When there is no variable can be used for correction
-        return mv_fisherz(mdata, X, Y, condition_set, sample_size)
+        return mv_fisherz(mdata, X, Y, condition_set)
 
     ## Learn regression models with test-wise deleted data
     involve_vars = var + W_indx
@@ -134,10 +132,12 @@ def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
 
     ## *********** Step 2 ***********
     # Get the data of the predictors, Ws
+    # The sample size of Ws is the same as the effective sample size
     Ws = Helper.get_predictor_ws(mdata[:, involve_vars], num_test_var=len(var), effective_sz=effective_sz)
 
     ## *********** Step 3 ***********
     # Generate the virtual data follows the full data distribution P(X, Y, S)
+    # The sample size of data_vir is the same as the effective sample size
     data_vir = Helper.gen_vir_data(regMs, rss, Ws, len(var), effective_sz)
 
     if len(var) > 2:
@@ -145,11 +145,11 @@ def mc_fisherz(mdata, skel, prt_m, X, Y, condition_set, sample_size):
     else:
         cond_set_bgn_0 = []
 
-    return mv_fisherz(data_vir, 0, 1, tuple(cond_set_bgn_0), effective_sz)
+    return mv_fisherz(data_vir, 0, 1, tuple(cond_set_bgn_0))
 
 
 def fisherz(data, X, Y, condition_set, correlation_matrix=None):
-    '''
+    """
     Perform an independence test using Fisher-Z's test
 
     Parameters
@@ -162,7 +162,7 @@ def fisherz(data, X, Y, condition_set, correlation_matrix=None):
     Returns
     -------
     p : the p-value of the test
-    '''
+    """
     if correlation_matrix is None:
         correlation_matrix = np.corrcoef(data.T)
     sample_size = data.shape[0]
@@ -191,17 +191,16 @@ def gsq(data, X, Y, conditioning_set, cardinalities=None):
 
 
 def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
-    '''by Haoyue@12/18/2021
+    """by Haoyue@12/18/2021
     Parameters
     ----------
     dataSXY: numpy.ndarray, in shape (|S|+2, n), where |S| is size of conditioning set (can be 0), n is sample size
              dataSXY.dtype = np.int64, and each row has values [0, 1, 2, ..., card_of_this_row-1]
     cardSXY: cardinalities of each row (each variable)
     G_sq: True if use G-sq, otherwise (False by default), use Chi_sq
-    '''
-
+    """
     def _Fill2DCountTable(dataXY, cardXY):
-        '''
+        """
         e.g. dataXY: the observed dataset contains 5 samples, on variable x and y they're
             x: 0 1 2 3 0
             y: 1 0 1 2 1
@@ -218,7 +217,7 @@ def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
             however some values may be missed.
             also in joint count, not every value in [0, cardX * cardY - 1] occurs.
             that's why we pass cardinalities in, and use `minlength=...` in bincount
-        '''
+        """
         cardX, cardY = cardXY
         xyIndexed = dataXY[0] * cardY + dataXY[1]
         xyJointCounts = np.bincount(xyIndexed, minlength=cardX * cardY).reshape(cardXY)
@@ -226,16 +225,14 @@ def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
         yMarginalCounts = np.sum(xyJointCounts, axis=0)
         return xyJointCounts, xMarginalCounts, yMarginalCounts
 
-    def _Fill3DCountTable(dataSXY, cardSXY):
+    def _Fill3DCountTableByBincount(dataSXY, cardSXY):
         cardX, cardY = cardSXY[-2:]
         cardS = np.prod(cardSXY[:-2])
-
         cardCumProd = np.ones_like(cardSXY)
         cardCumProd[:-1] = np.cumprod(cardSXY[1:][::-1])[::-1]
         SxyIndexed = np.dot(cardCumProd[None], dataSXY)[0]
 
         SxyJointCounts = np.bincount(SxyIndexed, minlength=cardS * cardX * cardY).reshape((cardS, cardX, cardY))
-
         SMarginalCounts = np.sum(SxyJointCounts, axis=(1, 2))
         SMarginalCountsNonZero = SMarginalCounts != 0
         SMarginalCounts = SMarginalCounts[SMarginalCountsNonZero]
@@ -245,8 +242,42 @@ def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
         SyJointCounts = np.sum(SxyJointCounts, axis=1)
         return SxyJointCounts, SMarginalCounts, SxJointCounts, SyJointCounts
 
+    def _Fill3DCountTableByUnique(dataSXY, cardSXY):
+        # Sometimes when the conditioning set contains many variables and each variable's cardinality is large
+        # e.g. consider an extreme case where
+        # S contains 7 variables and each's cardinality=20, then cardS = np.prod(cardSXY[:-2]) would be 1280000000
+        # i.e., there are 1280000000 different possible combinations of S,
+        #    so the SxyJointCounts array would be of size 1280000000 * cardX * cardY * np.int64,
+        #    i.e., ~3.73TB memory! (suppose cardX, cardX are also 20)
+        # However, samplesize is usually in 1k-100k scale, far less than cardS,
+        # i.e., not all (and actually only a very small portion of combinations of S appeared in data)
+        #    i.e., SMarginalCountsNonZero in _Fill3DCountTable_by_bincount is a very sparse array
+        # So when cardSXY is large, we first re-index S (skip the absent combinations) and then count XY table for each.
+        # See https://github.com/cmu-phil/causal-learn/pull/37.
+        cardX, cardY = cardSXY[-2:]
+        cardSs = cardSXY[:-2]
+
+        cardSsCumProd = np.ones_like(cardSs)
+        cardSsCumProd[:-1] = np.cumprod(cardSs[1:][::-1])[::-1]
+        SIndexed = np.dot(cardSsCumProd[None], dataSXY[:-2])[0]
+
+        uniqSIndices, inverseSIndices, SMarginalCounts = np.unique(SIndexed, return_counts=True, return_inverse=True)
+        cardS_reduced = len(uniqSIndices)
+        SxyIndexed = inverseSIndices * cardX * cardY + dataSXY[-2] * cardY + dataSXY[-1]
+        SxyJointCounts = np.bincount(SxyIndexed, minlength=cardS_reduced * cardX * cardY).reshape((cardS_reduced, cardX, cardY))
+
+        SxJointCounts = np.sum(SxyJointCounts, axis=2)
+        SyJointCounts = np.sum(SxyJointCounts, axis=1)
+        return SxyJointCounts, SMarginalCounts, SxJointCounts, SyJointCounts
+
+    def _Fill3DCountTable(dataSXY, cardSXY):
+        # about the threshold 1e5, see a rough performance example at:
+        # https://gist.github.com/MarkDana/e7d9663a26091585eb6882170108485e#file-count-unique-in-array-performance-md
+        if np.prod(cardSXY) < CONST_BINCOUNT_UNIQUE_THRESHOLD: return _Fill3DCountTableByBincount(dataSXY, cardSXY)
+        return _Fill3DCountTableByUnique(dataSXY, cardSXY)
+
     def _CalculatePValue(cTables, eTables):
-        '''
+        """
         calculate the rareness (pValue) of an observation from a given distribution with certain sample size.
 
         Let k, m, n be respectively the cardinality of S, x, y. if S=empty, k==1.
@@ -260,7 +291,7 @@ def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
 
         Returns: pValue (float in range (0, 1)), the larger pValue is (>alpha), the more independent.
         -------
-        '''
+        """
         eTables_zero_inds = eTables == 0
         eTables_zero_to_one = np.copy(eTables)
         eTables_zero_to_one[eTables_zero_inds] = 1  # for legal division
@@ -282,10 +313,10 @@ def chisq_or_gsq_test(dataSXY, cardSXY, G_sq=False):
         xyJointCounts, xMarginalCounts, yMarginalCounts = _Fill2DCountTable(dataSXY, cardSXY)
         xyExpectedCounts = np.outer(xMarginalCounts, yMarginalCounts) / dataSXY.shape[1]  # divide by sample size
         return _CalculatePValue(xyJointCounts[None], xyExpectedCounts[None])
+
     # else, S is not empty: conditioning
     SxyJointCounts, SMarginalCounts, SxJointCounts, SyJointCounts = _Fill3DCountTable(dataSXY, cardSXY)
     SxyExpectedCounts = SxJointCounts[:, :, None] * SyJointCounts[:, None, :] / SMarginalCounts[:, None, None]
-
     return _CalculatePValue(SxyJointCounts, SxyExpectedCounts)
 
 
@@ -300,7 +331,7 @@ def gsq_notoptimized(data, X, Y, conditioning_set):
 
 
 def chisq_or_gsq_test_notoptimized(data, X, Y, conditioning_set, G_sq=False):
-    '''
+    """
     Perform an independence test using chi-square test or G-square test
 
     Parameters
@@ -313,7 +344,7 @@ def chisq_or_gsq_test_notoptimized(data, X, Y, conditioning_set, G_sq=False):
     Returns
     -------
     p : the p-value of the test
-    '''
+    """
 
     # Step 1: Subset the data
     categories_list = [np.unique(data[:, i]) for i in
@@ -401,7 +432,7 @@ def cartesian_product(lists):
     return result
 
 
-def get_index_mv_rows(mvdata):
+def get_index_no_mv_rows(mvdata):
     nrow, ncol = np.shape(mvdata)
     bindxRows = np.ones((nrow,), dtype=bool)
     indxRows = np.array(list(range(nrow)))
@@ -409,10 +440,3 @@ def get_index_mv_rows(mvdata):
         bindxRows = np.logical_and(bindxRows, ~np.isnan(mvdata[:, i]))
     indxRows = indxRows[bindxRows]
     return indxRows
-
-
-def get_sub_correlation_matrix(mvdata):
-    indxRows = get_index_mv_rows(mvdata)
-    submatrix = np.corrcoef(mvdata[indxRows, :], rowvar=False)
-    sample_size = len(indxRows)
-    return submatrix, sample_size
